# lognews: A utility to grab Usenet article headers from a NNTP server
# and log them to a database.
#
# Copyright (c) Rayner Lucas 2004, 2010.
#
# This file is part of lognews.
#
# lognews is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lognews is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lognews.  If not, see <http://www.gnu.org/licenses/>.


# NewsCollector: module handling interactions with the NNTP server and
#                database server. Collects news from the NNTP server,
#                formats it, and logs it to the database.

# Module rfc822 is deprecated since Python 2.3. The email module should
# be used instead. However, the rfc822 module is still available for
# backward compatibility.
#import email
import rfc822
import nntplib
import signal
import string
import sys
import time

# Import the database module a little differently so we need fewer code
# changes if we switch to another PEP 249 compliant database module.
from MySQLdb import *


class NewsCollector:
	# Default values aren't really needed here - they'll always
	# be given values when the constructor is called.
	newshost = ''
	newsport = 119
	newsuser = ''
	newspass = ''
	dbhost = ''
	dbport = 3306
	dbuser = ''
	dbpass = ''
	dbname = ''
	nd = 0

	###### Constructor ######
	def __init__ (self, caller, nh, np, nu, npw, dbh, dbp, dbu, dbpw, dbn):
		signal.signal(signal.SIGINT, self.signal_handler)
		signal.signal(signal.SIGTERM, self.signal_handler)
		self.newshost = nh
		self.newsport = np
		self.newsuser = nu
		self.newspass = npw
		self.dbhost = dbh
		self.dbport = dbp
		self.dbuser = dbu
		self.dbpass = dbpw
		self.dbname = dbn
		self.nd = caller


	###### Public Methods ######

	# Perform checks for new articles and groups, and log new stuff to
	# the database.
	# Currently returns 1 for success and 0 for all failures.
	def DoChecks (self, lastchecktime):
		newsconn = nntplib.NNTP(self.newshost, self.newsport, self.newsuser, self.newspass)
		dbconn = connect(user=self.dbuser, passwd=self.dbpass, host=self.dbhost, port=self.dbport, db=self.dbname)

		# convert lastchecktime into date and time strings
		t = time.gmtime(lastchecktime)
		datestr = time.strftime("%y%m%d", t)
		timestr = time.strftime("%H%M%S", t)

		print 'datestr: ' + datestr
		print 'timestr: ' + timestr

		newgroups = self.getNewGroups(newsconn, datestr, timestr)
		for group in newgroups:
			print 'Found new group: ' + group + '.'
			self.addGroupToDB(dbconn, group)

		newarticles = self.getNewArticles(newsconn, datestr, timestr)
		for article in newarticles:
			print 'Found new article: ' + article + '.'
			headers = self.getArticleHeaders(newsconn, article)
			parsedhdrs = self.parseHeaders(headers)
			size = self.getArticleSize(newsconn, parsedhdrs["xref"][1], parsedhdrs["xref"][2])
			self.addArticleToDB(dbconn, parsedhdrs, size)

		return 1
		

	###### Private Methods ######

	# getNewGroups: return a list of new groups
	# conn: a nntplib.NNTP() connection object
	# datestr: the date of the last check for new groups, in YYMMDD format
	# timestr: the time of the last check for new groups, in HHMMSS format
	def getNewGroups (self, conn, datestr, timestr):
		response, groups = conn.newgroups(datestr, timestr)
		return groups


	# getNewArticles: return a list of message-ids of new articles
	# conn: a nntplib.NNTP() connection object
	# datestr: the date of the last check for new articles, in YYMMDD format
	# timestr: the time of the last check for new articles, in HHMMSS format
	def getNewArticles (self, conn, datestr, timestr):
		response, msgids = conn.newnews('*', datestr, timestr)
		return msgids


	# getArticleHeaders: return a list of headers for the specified article
	# conn: a nntplib.NNTP() connection object
	# msgid: The message-id of the article.
	def getArticleHeaders (self, conn, msgid):
		response, articlenum, id, headers = conn.head(msgid)
		return headers


	# getArticleSize: return the size of the specified article in bytes.
	#                 Requires the optional XOVER command.
	# conn: a nntplib.NNTP() connection object
	# group: a newsgroup to which the article was posted.
	# num: article's number in the newsgroup 'group'.
	def getArticleSize (self, conn, group, num):
		try:
			resp, count, first, last, name = conn.group(group)
			resp, list = conn.xover(num, num)
			if list:
				size = list[0][6]
			else:
				print 'Could not get size for message ' + group + ':' + str(num)
				size = 0
		except nntplib.NNTPTemporaryError:
			print 'Temporary NNTP error occurred while getting size for message ' + group + ':' + str(num) + '. Using size of 0.'
			size = 0
		return size


	# parseHeaders: returns a list of headers in a standard format, ready
	#               to be passed to addArticleToDB. Note: the header contents
	#               aren't sanitised here any more - to add them to the
	#               database, use parameter passing in cursor.execute() in
	#               order to make sure they're escaped correctly.
	# headers: The list of headers to be parsed.
	def parseHeaders (self, headers):
		parsedheaders = { "from": [], "path": [], "date": [], "tz": [],
			"subject": [], "msgid": [], "newsgroups": [], "lines": [],
			"references": [], "nntphost": [], "organization": [],
			"replyto": [], "sender": [], "approved": [], "control": [],
			"distrib": [], "expires": [], "exptz": [], "followupto": [],
			"xref": [] }

		# Glue headers with "continuation" lines (rfc1036, p2) together into
		# one string with embedded newlines. When we have all the lines of
		# the header, parse the line according to what kind of header it is
		# and store the results in the parsedheaders dict.
		i = 0
		while i < len(headers):
			thisline = headers[i]
			while (i+1 < len(headers)) and (headers[i+1][0] in (' ', '\t')):
				thisline += "\n" + headers[i+1]
				i += 1
			i += 1

			# Slight optimisation: Checks for required headers come first,
			# then checks for optional headers, commonest first.
			# Required headers:
			if thisline[0:6] == "From: ":
				parsedheaders["from"] = [ thisline[6:] ]
			elif thisline[0:6] == "Path: ":
				parsedheaders["path"] = self.splitPath(self.stripLineBreaks(thisline[6:]))
			elif thisline[0:6] == "Date: ":
				yr, mth, day, hr, min, sec, wd, yd, dst, tz = rfc822.parsedate_tz(thisline[6:])
				parsedheaders["date"] = [ time.strftime("%Y%m%d%H%M%S", (yr, mth, day, hr, min, sec, wd, yd, dst)) ]
				if tz == None:
					tz = 0
				parsedheaders["tz"] = [ `tz` ]
			elif thisline[0:9] == "Subject: ":
				parsedheaders["subject"] = [ thisline[9:] ]
			elif thisline[0:12] == "Message-ID: ":
				parsedheaders["msgid"] = [ thisline[12:] ]
			elif thisline[0:12] == "Newsgroups: ":
				tmpstr = self.stripLineBreaks(thisline[12:])
				parsedheaders["newsgroups"] = string.split(tmpstr, ",")
			elif thisline[0:6] == "Xref: ":
				# split into hostname and group:article-number pairs
				for str in string.split(thisline[6:], " "):
					# now split the group:article-number pairs
					for tmp in string.split(str, ':'):
						parsedheaders["xref"].append(tmp)

			# Non-required headers
			elif thisline[0:7] == "Lines: ":
				parsedheaders["lines"] = [ thisline[7:] ]
			elif thisline[0:12] == "References: ":
				tmpstr = self.stripLineBreaks(thisline[12:])
				parsedheaders["references"] = self.splitRefs(tmpstr)
			elif thisline[0:19] == "NNTP-Posting-Host: ":
				parsedheaders["nntphost"] = [ thisline[19:] ]
			elif thisline[0:14] == "Organization: ":
				parsedheaders["organization"] = [ thisline[14:] ]
			elif thisline[0:10] == "Reply-To: ":
				parsedheaders["replyto"] = [ thisline[10:] ]
			elif thisline[0:13] == "Followup-To: ":
				tmpstr = self.stripLineBreaks(thisline[13:])
				parsedheaders["followupto"] = string.split(tmpstr, ",")
			elif thisline[0:8] == "Sender: ":
				parsedheaders["sender"] = [ thisline[8:] ]
			elif thisline[0:10] == "Approved: ":
				parsedheaders["approved"] = [ thisline[10:] ]
			elif thisline[0:9] == "Control: ":
				parsedheaders["control"] = [ thisline[9:] ]
			elif thisline[0:14] == "Distribution: ":
				parsedheaders["distrib"] = string.split(thisline[14:], ",")
			elif thisline[0:9] == "Expires: ":
				yr, mth, day, hr, min, sec, wd, yd, dst, tz = rfc822.parsedate_tz(thisline[9:])
				parsedheaders["expires"] = [ time.strftime("%Y%m%d%H%M%S", (yr, mth, day, hr, min, sec, wd, yd, dst)) ]
				if tz == None:
					tz = 0
				parsedheaders["exptz"] = [ `tz` ]

		return parsedheaders


	# addGroupToDB: insert a new newsgroup into the database.
	# conn: A PEP 249 database connection object.
	# group: The name of the group to insert.
	def addGroupToDB (self, conn, group):
		c = conn.cursor()
		c.execute("INSERT INTO NEWSGROUP VALUES (%s)", (group,))


	# addArticleToDB: insert a new article into the database.
	# conn: A PEP 249 database connection object.
	# headers: The headers of the article to insert, in the format returned
	#          by the parseHeaders function.
	# size: The article's size in bytes, as returned by getArticleSize.
	def addArticleToDB (self, conn, headers, size):
		c = conn.cursor()

		for group in headers["newsgroups"]:
			c.execute("SELECT * FROM NEWSGROUP WHERE GROUPNAME = %s", (group,))
			if c.rowcount == 0:
				print 'Adding group ' + group
				self.addGroupToDB(conn, group)

		for dist in headers["distrib"]:
			c.execute("SELECT * FROM DISTRIBUTION WHERE DISTRIBUTIONNAME = %s", (dist,))
			if c.rowcount == 0:
				c.execute("INSERT INTO DISTRIBUTION VALUES (%s)", (dist,))

		for tz in headers["tz"]:
			c.execute("SELECT * FROM TIMEZONE WHERE OFFSET = %s", (tz,))
			if c.rowcount == 0:
				c.execute("INSERT INTO TIMEZONE VALUES (%s)", (tz,))

		for tz in headers["exptz"]:
			c.execute("SELECT * FROM TIMEZONE WHERE OFFSET = %s", (tz,))
			if c.rowcount == 0:
				c.execute("INSERT INTO TIMEZONE VALUES (%s)", (tz,))

		for server in headers["path"]:
			c.execute("SELECT * FROM NEWSSERVER WHERE SERVERNAME = %s", (server,))
			if c.rowcount == 0:
				c.execute("INSERT INTO NEWSSERVER VALUES (%s)", (server,))

		# Build the INSERT statement for the ARTICLE table.
		# If an entry in the article_fields dict is set to None because that
		# header did not exist, that field will become a NULL in the database.
		article_fields = {}
		for field in ['msgid', 'from', 'date', 'tz', 'subject', 'expires', 'exptz', 'replyto', 'sender', 'control', 'approved', 'lines', 'organization', 'nntphost']:
			if headers[field]:
				article_fields[field] = headers[field][0]
			else:
				article_fields[field] = None
		c.execute("INSERT INTO ARTICLE (MESSAGEID, ARTICLEFROM, ARTICLEDATE, DATETZ, SUBJECT, EXPIRES, EXPTZ, REPLYTO, SENDER, CONTROL, APPROVED, NUMLINES, ORGANIZATION, SIZE, NNTPPOSTINGHOST) VALUES (%s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s, %s)", (article_fields['msgid'], article_fields['from'], article_fields['date'], article_fields['tz'], article_fields['subject'], article_fields['expires'], article_fields['exptz'], article_fields['replyto'], article_fields['sender'], article_fields['control'], article_fields['approved'], article_fields['lines'], article_fields['organization'], str(size), article_fields['nntphost']))

		id = c.lastrowid
		id = str(id)

		for ref in headers["references"]:
			c.execute("SELECT * FROM REFERENCE WHERE REFERENCER = %s AND REFERENCED = %s", (id, ref))
			if c.rowcount == 0:
				c.execute("INSERT INTO REFERENCE VALUES (%s, %s)", (id, ref))

		for i in range(len(headers["path"])):
			c.execute("INSERT INTO NEWSSERVERARTICLE VALUES (%s, %s, %s)", (id, headers["path"][i], str(i)))

		for dist in headers["distrib"]:
			c.execute("INSERT INTO ARTICLEDISTRIBUTION VALUES (%s, %s)", (id, dist))

		for group in headers["newsgroups"]:
			c.execute("SELECT * FROM NEWSGROUPARTICLE WHERE ARTICLEID = %s AND GROUPNAME = %s", (id, group))
			if c.rowcount == 0:
				c.execute("INSERT INTO NEWSGROUPARTICLE VALUES (%s, %s)", (id, group))

		for group in headers["followupto"]:
			c.execute("INSERT INTO FOLLOWUPNEWSGROUP VALUES (%s, %s)", (id, group))


	# splitPath: return an array of strings (hostnames) given the contents
	# of a Path header. Hostnames may be separated by sequences of one
	# or more punctuation characters. A punctuation character is any
	# character that is not a letter, digit, full stop or hyphen (RFC1036,
	# section 2.1.6)
	# path: The contents of the Path header to be parsed.
	def splitPath (self, path):
		hosts = []
		i = 0
		path = string.strip(path)
		while i < len(path):
			while i < len(path) and not (path[i].isalnum() or path[i] in ('-', '.')):
				i += 1
			start = i
			while i < len(path) and (path[i].isalnum() or path[i] in ('-', '.')):
				i += 1
			end = i
			hosts.append(path[start:end])
		return hosts


	# splitRefs: return an array of message-IDs given the contents of
	#            of a References header. Message-IDs may be separated by
	#            any number of blanks, tabs and newlines (this deals with
	#            multi-line References headers).
	# hdr: The contents of the References header to be parsed.
	def splitRefs(self, hdr):
		refs = []
		i = 0
		hdr = string.strip(hdr)
		while i < len(hdr):
			while i < len(hdr) and hdr[i] in (' ', '\t', '\n'):
				i += 1
			start = i
			while i < len(hdr) and hdr[i] not in (' ', '\t', '\n'):
				i += 1
			end = i
			refs.append(hdr[start:end])
		return refs


	# stripLineBreaks: Return the input string with newlines and any
	#                  blanks or tabs immediately following them removed.
	# str: The input string.
	def stripLineBreaks(self, str):
		newstr = ''
		i = 0
		while i < len(str):
			if str[i] != '\n':
				newstr += str[i]
				i += 1
			else:
				while i < len(str) and str[i] in (' ', '\t', '\n'):
					i += 1
		return newstr


	# signal_handler: trap external signals sent to the program.
	#
	# Signals currently handled are:
	# SIGTERM: call cleanup(), then exit successfully.
	# SIGINT: call cleanup(), then exit successfully.
	def signal_handler(self, signal, frame):
		print 'SIGINT or SIGTERM caught in lognews newscollector, exiting...'
		self.nd.cleanup()
		sys.exit(0)

